@@ -32,7 +32,6 @@
 #include <signal.h>
 #include <stdint.h>
 #include <errno.h>
-#include <syslog.h>
 #include "xs.h"
 #include "list.h"
 #include "utils.h"
@@ -80,12 +79,34 @@ struct xs_handle {
 
 	/* One request at a time. */
 	pthread_mutex_t request_mutex;
+
+	/* Lock discipline:
+	 *  Only holder of the request lock may write to h->fd.
+	 *  Only holder of the request lock may access read_thr_exists.
+	 *  If read_thr_exists==0, only holder of request lock may read h->fd;
+	 *  If read_thr_exists==1, only the read thread may read h->fd.
+	 *  Only holder of the reply lock may access reply_list.
+	 *  Only holder of the watch lock may access watch_list.
+	 * Lock hierarchy:
+	 *  The order in which to acquire locks is
+	 *     request_mutex
+	 *     reply_mutex
+	 *     watch_mutex
+	 */
 };
 
 #define mutex_lock(m)		pthread_mutex_lock(m)
 #define mutex_unlock(m)		pthread_mutex_unlock(m)
 #define condvar_signal(c)	pthread_cond_signal(c)
-#define condvar_wait(c,m,hnd)	pthread_cond_wait(c,m)
+#define condvar_wait(c,m)	pthread_cond_wait(c,m)
+#define cleanup_push(f, a)	\
+    pthread_cleanup_push((void (*)(void *))(f), (void *)(a))
+/*
+ * Some definitions of pthread_cleanup_pop() are a macro starting with an
+ * end-brace. GCC then complains if we immediately precede that with a label.
+ * Hence we insert a dummy statement to appease the compiler in this situation.
+ */
+#define cleanup_pop(run)        ((void)0); pthread_cleanup_pop(run)
 
 #define read_thread_exists(h)	(h->read_thr_exists)
 
@@ -104,8 +125,9 @@ struct xs_handle {
 #define mutex_lock(m)		((void)0)
 #define mutex_unlock(m)		((void)0)
 #define condvar_signal(c)	((void)0)
-#define condvar_wait(c,m,hnd)	read_message(hnd)
-
+#define condvar_wait(c,m)	((void)0)
+#define cleanup_push(f, a)	((void)0)
+#define cleanup_pop(run)	((void)0)
 #define read_thread_exists(h)	(0)
 
 #endif
@@ -160,12 +182,15 @@ error:
 	return -1;
 }
 
-static int get_dev(const char *connect_to)
+static int get_dev(const char *connect_to, unsigned long flags)
 {
-	return open(connect_to, O_RDWR);
+	if (flags & XS_OPEN_READONLY)
+		return open(connect_to, O_RDONLY);
+	else
+		return open(connect_to, O_RDWR);
 }
 
-static struct xs_handle *get_handle(const char *connect_to)
+static struct xs_handle *get_handle(const char *connect_to, unsigned long flags)
 {
 	struct stat buf;
 	struct xs_handle *h = NULL;
@@ -177,7 +202,7 @@ static struct xs_handle *get_handle(const char *connect_to)
 	if (S_ISSOCK(buf.st_mode))
 		fd = get_socket(connect_to);
 	else
-		fd = get_dev(connect_to);
+		fd = get_dev(connect_to, flags);
 
 	if (fd == -1)
 		return NULL;
@@ -215,34 +240,42 @@ static struct xs_handle *get_handle(const char *connect_to)
 
 struct xs_handle *xs_daemon_open(void)
 {
-	return get_handle(xs_daemon_socket());
+	return xs_open(0);
 }
 
 struct xs_handle *xs_daemon_open_readonly(void)
 {
-	return get_handle(xs_daemon_socket_ro());
+	return xs_open(XS_OPEN_READONLY);
 }
 
 struct xs_handle *xs_domain_open(void)
 {
-	return get_handle(xs_domain_dev());
+	return xs_open(0);
 }
 
-void xs_daemon_close(struct xs_handle *h)
+struct xs_handle *xs_open(unsigned long flags)
 {
-	struct xs_stored_msg *msg, *tmsg;
+	struct xs_handle *xsh = NULL;
 
-	mutex_lock(&h->request_mutex);
-	mutex_lock(&h->reply_mutex);
-	mutex_lock(&h->watch_mutex);
-
-#ifdef USE_PTHREAD
-	if (h->read_thr_exists) {
-		/* XXX FIXME: May leak an unpublished message buffer. */
-		pthread_cancel(h->read_thr);
-		pthread_join(h->read_thr, NULL);
+	flags |= (getenv("XS_OPEN_READONLY") ? XS_OPEN_READONLY : 0);
+	flags |= (getenv("XS_OPEN_SOCKETONLY") ? XS_OPEN_SOCKETONLY : 0);
+	flags |= (getenv("XS_OPEN_XENBUSONLY") ? XS_OPEN_XENBUSONLY : 0);
+       
+	if ((flags & XS_OPEN_SOCKETONLY) || !(flags & XS_OPEN_XENBUSONLY)) {
+		if (flags & XS_OPEN_READONLY)
+			xsh = get_handle(xs_daemon_socket_ro(), flags);
+		else
+			xsh = get_handle(xs_daemon_socket(), flags);
 	}
-#endif
+
+	if (!xsh && ((flags & XS_OPEN_XENBUSONLY) || !(flags & XS_OPEN_SOCKETONLY)))
+		xsh = get_handle(xs_domain_dev(), flags);
+
+	return xsh;
+}
+
+static void close_free_msgs(struct xs_handle *h) {
+	struct xs_stored_msg *msg, *tmsg;
 
 	list_for_each_entry_safe(msg, tmsg, &h->reply_list, list) {
 		free(msg->body);
@@ -253,21 +286,53 @@ void xs_daemon_close(struct xs_handle *h)
 		free(msg->body);
 		free(msg);
 	}
+}
 
-	mutex_unlock(&h->request_mutex);
-	mutex_unlock(&h->reply_mutex);
-	mutex_unlock(&h->watch_mutex);
-
+static void close_fds_free(struct xs_handle *h) {
 	if (h->watch_pipe[0] != -1) {
 		close(h->watch_pipe[0]);
 		close(h->watch_pipe[1]);
 	}
 
-	close(h->fd);
-
+        close(h->fd);
+        
 	free(h);
 }
 
+void xs_daemon_destroy_postfork(struct xs_handle *h)
+{
+        close_free_msgs(h);
+        close_fds_free(h);
+}
+
+void xs_daemon_close(struct xs_handle *h)
+{
+#ifdef USE_PTHREAD
+	if (h->read_thr_exists) {
+		pthread_cancel(h->read_thr);
+		pthread_join(h->read_thr, NULL);
+	}
+#endif
+
+	mutex_lock(&h->request_mutex);
+	mutex_lock(&h->reply_mutex);
+	mutex_lock(&h->watch_mutex);
+
+        close_free_msgs(h);
+
+	mutex_unlock(&h->request_mutex);
+	mutex_unlock(&h->reply_mutex);
+	mutex_unlock(&h->watch_mutex);
+
+        close_fds_free(h);
+}
+
+void xs_close(struct xs_handle* xsh)
+{
+	if (xsh)
+		xs_daemon_close(xsh);
+}
+
 static bool read_all(int fd, void *data, unsigned int len)
 {
 	while (len) {
@@ -316,21 +381,20 @@ static void *read_reply(
 
 	read_from_thread = read_thread_exists(h);
 
-#ifdef USE_PTHREAD
 	/* Read from comms channel ourselves if there is no reader thread. */
 	if (!read_from_thread && (read_message(h) == -1))
 		return NULL;
-#endif
 
 	mutex_lock(&h->reply_mutex);
-	while (list_empty(&h->reply_list) && (!read_from_thread || read_thread_exists(h)))
-		condvar_wait(&h->reply_condvar, &h->reply_mutex, h);
-	if (read_from_thread && !read_thread_exists(h)) {
+#ifdef USE_PTHREAD
+	while (list_empty(&h->reply_list) && read_from_thread && h->fd != -1)
+		condvar_wait(&h->reply_condvar, &h->reply_mutex);
+#endif
+	if (list_empty(&h->reply_list)) {
 		mutex_unlock(&h->reply_mutex);
 		errno = EINVAL;
 		return NULL;
 	}
-	assert(!list_empty(&h->reply_list));
 	msg = list_top(&h->reply_list, struct xs_stored_msg, list);
 	list_del(&msg->list);
 	assert(list_empty(&h->reply_list));
@@ -612,14 +676,36 @@ bool xs_watch(struct xs_handle *h, const char *path, const char *token)
 	struct iovec iov[2];
 
 #ifdef USE_PTHREAD
+#define READ_THREAD_STACKSIZE (16 * 1024)
+
 	/* We dynamically create a reader thread on demand. */
 	mutex_lock(&h->request_mutex);
 	if (!h->read_thr_exists) {
-		if (pthread_create(&h->read_thr, NULL, read_thread, h) != 0) {
+		sigset_t set, old_set;
+		pthread_attr_t attr;
+
+		if (pthread_attr_init(&attr) != 0) {
+			mutex_unlock(&h->request_mutex);
+			return false;
+		}
+		if (pthread_attr_setstacksize(&attr, READ_THREAD_STACKSIZE) != 0) {
+			pthread_attr_destroy(&attr);
+			mutex_unlock(&h->request_mutex);
+			return false;
+		}
+
+		sigfillset(&set);
+		pthread_sigmask(SIG_SETMASK, &set, &old_set);
+
+		if (pthread_create(&h->read_thr, &attr, read_thread, h) != 0) {
+			pthread_sigmask(SIG_SETMASK, &old_set, NULL);
+			pthread_attr_destroy(&attr);
 			mutex_unlock(&h->request_mutex);
 			return false;
 		}
 		h->read_thr_exists = 1;
+		pthread_sigmask(SIG_SETMASK, &old_set, NULL);
+		pthread_attr_destroy(&attr);
 	}
 	mutex_unlock(&h->request_mutex);
 #endif
@@ -645,10 +731,25 @@ char **xs_read_watch(struct xs_handle *h, unsigned int *num)
 
 	mutex_lock(&h->watch_mutex);
 
-	/* Wait on the condition variable for a watch to fire. */
-	while (list_empty(&h->watch_list) && read_thread_exists(h))
-		condvar_wait(&h->watch_condvar, &h->watch_mutex, h);
-	if (!read_thread_exists(h)) {
+#ifdef USE_PTHREAD
+	/* Wait on the condition variable for a watch to fire.
+	 * If the reader thread doesn't exist yet, then that's because
+	 * we haven't called xs_watch.	Presumably the application
+	 * will do so later; in the meantime we just block.
+	 */
+	while (list_empty(&h->watch_list) && h->fd != -1)
+		condvar_wait(&h->watch_condvar, &h->watch_mutex);
+#else /* !defined(USE_PTHREAD) */
+	/* Read from comms channel ourselves if there are no threads
+	 * and therefore no reader thread. */
+
+	assert(!read_thread_exists(h)); /* not threadsafe but worth a check */
+	if ((read_message(h) == -1))
+		return NULL;
+
+#endif /* !defined(USE_PTHREAD) */
+
+	if (list_empty(&h->watch_list)) {
 		mutex_unlock(&h->watch_mutex);
 		errno = EINVAL;
 		return NULL;
@@ -871,67 +972,59 @@ char *xs_debug_command(struct xs_handle *h, const char *cmd,
 
 static int read_message(struct xs_handle *h)
 {
+	/* IMPORTANT: It is forbidden to call this function without
+	 * acquiring the request lock and checking that h->read_thr_exists
+	 * is false.  See "Lock discipline" in struct xs_handle, above. */
+         
 	struct xs_stored_msg *msg = NULL;
 	char *body = NULL;
-	int saved_errno;
+	int saved_errno = 0;
+	int ret = -1;
 
 	/* Allocate message structure and read the message header. */
 	msg = malloc(sizeof(*msg));
-	if (msg == NULL) {
-		openlog("libxenstore", LOG_CONS|LOG_PID, LOG_DAEMON);
-		syslog(LOG_DAEMON|LOG_EMERG,
-		       "read_message malloc msg failed: %m");
-		goto error;
-	}
-	if (!read_all(h->fd, &msg->hdr, sizeof(msg->hdr))) {
-		openlog("libxenstore", LOG_CONS|LOG_PID, LOG_DAEMON);
-		syslog(LOG_DAEMON|LOG_EMERG,
-		       "read_message read_all header failed: %m");
+	if (msg == NULL)
 		goto error;
+	cleanup_push(free, msg);
+	if (!read_all(h->fd, &msg->hdr, sizeof(msg->hdr))) { /* Cancellation point */
+		saved_errno = errno;
+		goto error_freemsg;
 	}
 
 	/* Allocate and read the message body. */
 	body = msg->body = malloc(msg->hdr.len + 1);
-	if (body == NULL) {
-		openlog("libxenstore", LOG_CONS|LOG_PID, LOG_DAEMON);
-		syslog(LOG_DAEMON|LOG_EMERG,
-		       "read_message malloc body %d failed: %m",
-		       msg->hdr.len + 1);
-		goto error;
-	}
-	if (!read_all(h->fd, body, msg->hdr.len)) {
-		openlog("libxenstore", LOG_CONS|LOG_PID, LOG_DAEMON);
-		syslog(LOG_DAEMON|LOG_EMERG,
-		       "read_message read_all body %d failed: %m",
-		       msg->hdr.len);
-		goto error;
+	if (body == NULL)
+		goto error_freemsg;
+	cleanup_push(free, body);
+	if (!read_all(h->fd, body, msg->hdr.len)) { /* Cancellation point */
+		saved_errno = errno;
+		goto error_freebody;
 	}
+
 	body[msg->hdr.len] = '\0';
 
 	if (msg->hdr.type == XS_WATCH_EVENT) {
 		mutex_lock(&h->watch_mutex);
+		cleanup_push(pthread_mutex_unlock, &h->watch_mutex);
 
 		/* Kick users out of their select() loop. */
 		if (list_empty(&h->watch_list) &&
 		    (h->watch_pipe[1] != -1))
-			while (write(h->watch_pipe[1], body, 1) != 1)
+			while (write(h->watch_pipe[1], body, 1) != 1) /* Cancellation point */
 				continue;
 
 		list_add_tail(&msg->list, &h->watch_list);
 
 		condvar_signal(&h->watch_condvar);
 
-		mutex_unlock(&h->watch_mutex);
+		cleanup_pop(1);
 	} else {
 		mutex_lock(&h->reply_mutex);
 
 		/* There should only ever be one response pending! */
 		if (!list_empty(&h->reply_list)) {
 			mutex_unlock(&h->reply_mutex);
-			openlog("libxenstore", LOG_CONS|LOG_PID, LOG_DAEMON);
-			syslog(LOG_DAEMON|LOG_EMERG,
-			       "read_message list not empty");
-			goto error;
+			goto error_freebody;
 		}
 
 		list_add_tail(&msg->list, &h->reply_list);
@@ -940,35 +1033,43 @@ static int read_message(struct xs_handle *h)
 		mutex_unlock(&h->reply_mutex);
 	}
 
-	return 0;
+	ret = 0;
 
- error:
-	saved_errno = errno;
-	free(msg);
-	free(body);
+error_freebody:
+	cleanup_pop(ret == -1);
+error_freemsg:
+	cleanup_pop(ret == -1);
+error:
 	errno = saved_errno;
-	return -1;
+
+	return ret;
 }
 
 #ifdef USE_PTHREAD
 static void *read_thread(void *arg)
 {
 	struct xs_handle *h = arg;
+	int fd;
 
 	while (read_message(h) != -1)
 		continue;
 
-	/* Kick anyone waiting for a reply */
-	pthread_mutex_lock(&h->request_mutex);
-	h->read_thr_exists = 0;
-	pthread_mutex_unlock(&h->request_mutex);
+	/* An error return from read_message leaves the socket in an undefined
+	 * state; we might have read only the header and not the message after
+	 * it, or (more commonly) the other end has closed the connection.
+	 * Since further communication is unsafe, close the socket.
+	 */
+	fd = h->fd;
+	h->fd = -1;
+	close(fd);
 
+	/* wake up all waiters */
 	pthread_mutex_lock(&h->reply_mutex);
-	pthread_cond_signal(&h->reply_condvar);
+	pthread_cond_broadcast(&h->reply_condvar);
 	pthread_mutex_unlock(&h->reply_mutex);
 
 	pthread_mutex_lock(&h->watch_mutex);
-	pthread_cond_signal(&h->watch_condvar);
+	pthread_cond_broadcast(&h->watch_condvar);
 	pthread_mutex_unlock(&h->watch_mutex);
 
 	return NULL;
