@@ -9,20 +9,23 @@
 #include <xen/shared.h>
 #include <public/xen.h>
 #include <public/domctl.h>
+#include <public/sysctl.h>
 #include <public/vcpu.h>
 #include <public/xsm/acm.h>
 #include <xen/time.h>
 #include <xen/timer.h>
 #include <xen/grant_table.h>
 #include <xen/rangeset.h>
-#include <asm/domain.h>
+#include <xen/domain.h>
 #include <xen/xenoprof.h>
 #include <xen/rcupdate.h>
 #include <xen/irq.h>
 #include <xen/mm.h>
-#if _MTC_LOCKSTEP
-    #include <mtc/lockstep.h>
-#endif
+#include <xen/tasklet.h>
+#include <public/mem_event.h>
+#include <xen/cpumask.h>
+#include <xen/nodemask.h>
+#include <xen/multicall.h>
 
 #ifdef CONFIG_COMPAT
 #include <compat/vcpu.h>
@@ -61,7 +64,11 @@ struct evtchn
             u16            remote_port;
             struct domain *remote_dom;
         } interdomain; /* state == ECS_INTERDOMAIN */
-        u16 pirq;      /* state == ECS_PIRQ */
+        struct {
+            u16            irq;
+            u16            next_port;
+            u16            prev_port;
+        } pirq;        /* state == ECS_PIRQ */
         u16 virq;      /* state == ECS_VIRQ */
     } u;
 #ifdef FLASK_ENABLE
@@ -69,8 +76,11 @@ struct evtchn
 #endif
 };
 
-int  evtchn_init(struct domain *d);
-void evtchn_destroy(struct domain *d);
+int  evtchn_init(struct domain *d); /* from domain_create */
+void evtchn_destroy(struct domain *d); /* from domain_kill */
+void evtchn_destroy_final(struct domain *d); /* from complete_domain_destroy */
+
+struct waitqueue_vcpu;
 
 struct vcpu 
 {
@@ -116,27 +126,23 @@ struct vcpu
     bool_t           is_initialised;
     /* Currently running on a CPU? */
     bool_t           is_running;
-    /* MCE callback pending for this VCPU? */
-    bool_t           mce_pending;
-    /* NMI callback pending for this VCPU? */
-    bool_t           nmi_pending;
+    /* VCPU should wake fast (do not deep sleep the CPU). */
+    bool_t           is_urgent;
 
-    /* Higher priorized traps may interrupt lower priorized traps,
-     * lower priorized traps wait until higher priorized traps finished.
-     * Note: This concept is known as "system priority level" (spl)
-     * in the UNIX world. */
-    uint16_t         old_trap_priority;
-    uint16_t         trap_priority;
+#ifdef VCPU_TRAP_LAST
 #define VCPU_TRAP_NONE    0
-#define VCPU_TRAP_NMI     1
-#define VCPU_TRAP_MCE     2
+    struct {
+        bool_t           pending;
+        uint8_t          old_mask;
+    }                async_exception_state[VCPU_TRAP_LAST];
+#define async_exception_state(t) async_exception_state[(t)-1]
+    uint8_t          async_exception_mask;
+#endif
 
     /* Require shutdown to be deferred for some asynchronous operation? */
     bool_t           defer_shutdown;
     /* VCPU is paused following shutdown request (d->is_shutting_down)? */
     bool_t           paused_for_shutdown;
-    /* VCPU affinity is temporarily locked from controller changes? */
-    bool_t           affinity_locked;
 
     /*
      * > 0: a single port is being polled;
@@ -145,6 +151,9 @@ struct vcpu
      */
     int              poll_evtchn;
 
+    /* (over-)protected by ->domain->event_lock */
+    int              pirq_evtchn_head;
+
     unsigned long    pause_flags;
     atomic_t         pause_count;
 
@@ -160,14 +169,15 @@ struct vcpu
     /* Bitmask of CPUs which are holding onto this VCPU's state. */
     cpumask_t        vcpu_dirty_cpumask;
 
-    struct arch_vcpu arch;
+    /* Tasklet for continue_hypercall_on_cpu(). */
+    struct tasklet   continue_hypercall_tasklet;
 
-#if _MTC_LOCKSTEP
-    LOCKSTEP_VCPU * lockstep;         /* ptr to Lockstep Assist vcpu descriptor extension */
-    uint32          vmexit_num;       /* vm-exit number */
-    uint32          tos_exit_num;     /* TOS-exit number */
-    uint32          exit_num_log;     /* vm-exit or tos-exit number of last log */
-#endif
+    /* Multicall information. */
+    struct mc_state  mc_state;
+
+    struct waitqueue_vcpu *waitqueue_vcpu;
+
+    struct arch_vcpu arch;
 };
 
 /* Per-domain lock can be recursively acquired in fault handlers. */
@@ -175,9 +185,20 @@ struct vcpu
 #define domain_unlock(d) spin_unlock_recursive(&(d)->domain_lock)
 #define domain_is_locked(d) spin_is_locked(&(d)->domain_lock)
 
-
-/* Number of supported watchdog timers per domain */
-#define NR_WATCHDOG_TIMERS 2
+/* Memory event */
+struct mem_event_domain
+{
+    /* ring lock */
+    spinlock_t ring_lock;
+    /* shared page */
+    mem_event_shared_page_t *shared_page;
+    /* shared ring page */
+    void *ring_page;
+    /* front-end ring */
+    mem_event_front_ring_t front_ring;
+    /* event channel port (vcpu0 only) */
+    int xen_port;
+};
 
 typedef struct xen_domctl_runstate_info domain_runstate_info_t;
 
@@ -190,14 +211,18 @@ struct domain
     spinlock_t       domain_lock;
 
     spinlock_t       page_alloc_lock; /* protects all the following fields  */
-    struct page_list_head page_list;  /* linked list, of size tot_pages     */
+    struct page_list_head page_list;  /* linked list */
     struct page_list_head xenpage_list; /* linked list (size xenheap_pages) */
     unsigned int     tot_pages;       /* number of pages currently possesed */
     unsigned int     max_pages;       /* maximum value for tot_pages        */
+    atomic_t         shr_pages;       /* number of shared pages             */
     unsigned int     xenheap_pages;   /* # pages allocated from Xen heap    */
 
+    unsigned int     max_vcpus;
+
     /* Scheduling. */
     void            *sched_priv;    /* scheduler-specific data */
+    struct cpupool  *cpupool;
 
     struct domain   *next_in_list;
     struct domain   *next_in_hashbucket;
@@ -216,8 +241,9 @@ struct domain
      * domain's event-channel spinlock. Read accesses can also synchronise on 
      * the lock, but races don't usually matter.
      */
-    u16              pirq_to_evtchn[NR_IRQS];
-    DECLARE_BITMAP(pirq_mask, NR_IRQS);
+    unsigned int     nr_pirqs;
+    u16             *pirq_to_evtchn;
+    unsigned long   *pirq_mask;
 
     /* I/O capabilities (access to IRQs and memory-mapped I/O). */
     struct rangeset *iomem_caps;
@@ -241,13 +267,16 @@ struct domain
     bool_t           is_pinned;
 
     /* Are any VCPUs polling event channels (SCHEDOP_poll)? */
+#if MAX_VIRT_CPUS <= BITS_PER_LONG
     DECLARE_BITMAP(poll_mask, MAX_VIRT_CPUS);
+#else
+    unsigned long   *poll_mask;
+#endif
 
     /* Guest has shut down (inc. reason code)? */
     spinlock_t       shutdown_lock;
     bool_t           is_shutting_down; /* in process of shutting down? */
     bool_t           is_shut_down;     /* fully shut down? */
-    bool_t           has_shutdown_code; /* shutdown_code has been set? */
     int              shutdown_code;
 
     /* If this is not 0, send suspend notification here instead of
@@ -260,7 +289,7 @@ struct domain
 
     atomic_t         refcnt;
 
-    struct vcpu *vcpu[MAX_VIRT_CPUS];
+    struct vcpu    **vcpu;
 
     /* Bitmask of CPUs which are holding onto this domain's state. */
     cpumask_t        domain_dirty_cpumask;
@@ -277,8 +306,10 @@ struct domain
     int32_t time_offset_seconds;
 
     /* Domain watchdog. */
+#define NR_DOMAIN_WATCHDOG_TIMERS 2
     spinlock_t watchdog_lock;
-    struct timer watchdog_timer[NR_WATCHDOG_TIMERS];
+    uint32_t watchdog_inuse_map;
+    struct timer watchdog_timer[NR_DOMAIN_WATCHDOG_TIMERS];
 
     struct rcu_head rcu;
 
@@ -288,18 +319,28 @@ struct domain
      */
     spinlock_t hypercall_deadlock_mutex;
 
-    /* VRAM dirty support. */
-    struct sh_dirty_vram *dirty_vram;
+    /* transcendent memory, auto-allocated on first tmem op by each domain */
+    void *tmem;
 
+    struct lock_profile_qhead profile_head;
+
+    /* Non-migratable and non-restoreable? */
+    bool_t disable_migrate;
+
+    /* Memory paging support */
+    struct mem_event_domain mem_event;
+
+    /* Currently computed from union of all vcpu cpu-affinity masks. */
+    nodemask_t node_affinity;
+    unsigned int last_alloc_node;
+    spinlock_t node_affinity_lock;
+
+    /* Domain runstates */
     spinlock_t runstate_lock;
     atomic_t runstate_missed_changes;
     domain_runstate_info_t runstate;
 
     unsigned int cores_per_socket;
-#if _MTC_LOCKSTEP
-    /* ptr to Lockastep Assist domain descriptor extension */
-    LOCKSTEP_DOMAIN * lockstep;
-#endif
 };
 
 struct domain_setup_info
@@ -325,9 +366,12 @@ struct domain_setup_info
     unsigned long symtab_len;
 };
 
+/* Protect updates/reads (resp.) of domain_list and domain_hash. */
+extern spinlock_t domlist_update_lock;
+extern rcu_read_lock_t domlist_read_lock;
+
 extern struct vcpu *idle_vcpu[NR_CPUS];
-#define IDLE_DOMAIN_ID   (0x7FFFU)
-#define is_idle_domain(d) ((d)->domain_id == IDLE_DOMAIN_ID)
+#define is_idle_domain(d) ((d)->domain_id == DOMID_IDLE)
 #define is_idle_vcpu(v)   (is_idle_domain((v)->domain))
 
 #define DOMAIN_DESTROYED (1<<31) /* assumes atomic_t is >= 32 bits */
@@ -363,13 +407,7 @@ static inline void get_knownalive_domain(struct domain *d)
     ASSERT(!(atomic_read(&d->refcnt) & DOMAIN_DESTROYED));
 }
 
-/* Obtain a reference to the currently-running domain. */
-static inline struct domain *get_current_domain(void)
-{
-    struct domain *d = current->domain;
-    get_knownalive_domain(d);
-    return d;
-}
+void domain_update_node_affinity(struct domain *d);
 
 struct domain *domain_create(
     domid_t domid, unsigned int domcr_flags, ssidref_t ssidref);
@@ -386,6 +424,9 @@ struct domain *domain_create(
  /* DOMCRF_dummy: Create a dummy domain (not scheduled; not on domain list) */
 #define _DOMCRF_dummy         3
 #define DOMCRF_dummy          (1U<<_DOMCRF_dummy)
+ /* DOMCRF_oos_off: dont use out-of-sync optimization for shadow page tables */
+#define _DOMCRF_oos_off         4
+#define DOMCRF_oos_off          (1U<<_DOMCRF_oos_off)
 
 /*
  * rcu_lock_domain_by_id() is more efficient than get_domain_by_id().
@@ -403,21 +444,30 @@ struct domain *rcu_lock_domain_by_id(domid_t dom);
  */
 int rcu_lock_target_domain_by_id(domid_t dom, struct domain **d);
 
+/*
+ * As rcu_lock_target_domain_by_id(), but will fail EPERM rather than resolve
+ * to local domain. Successful return always resolves to a remote domain that
+ * the local domain is privileged to control.
+ */
+int rcu_lock_remote_target_domain_by_id(domid_t dom, struct domain **d);
+
 /* Finish a RCU critical region started by rcu_lock_domain_by_id(). */
 static inline void rcu_unlock_domain(struct domain *d)
 {
-    rcu_read_unlock(&domlist_read_lock);
+    if ( d != current->domain )
+        rcu_read_unlock(&domlist_read_lock);
 }
 
 static inline struct domain *rcu_lock_domain(struct domain *d)
 {
-    rcu_read_lock(d);
+    if ( d != current->domain )
+        rcu_read_lock(d);
     return d;
 }
 
 static inline struct domain *rcu_lock_current_domain(void)
 {
-    return rcu_lock_domain(current->domain);
+    return /*rcu_lock_domain*/(current->domain);
 }
 
 struct domain *get_domain_by_id(domid_t dom);
@@ -457,7 +507,9 @@ int  sched_init_vcpu(struct vcpu *v, unsigned int processor);
 void sched_destroy_vcpu(struct vcpu *v);
 int  sched_init_domain(struct domain *d);
 void sched_destroy_domain(struct domain *d);
+int sched_move_domain(struct domain *d, struct cpupool *c);
 long sched_adjust(struct domain *, struct xen_domctl_scheduler_op *);
+long sched_adjust_global(struct xen_sysctl_scheduler_op *);
 int  sched_id(void);
 void sched_tick_suspend(void);
 void sched_tick_resume(void);
@@ -472,6 +524,9 @@ void vcpu_sleep_sync(struct vcpu *d);
  */
 void sync_vcpu_execstate(struct vcpu *v);
 
+/* As above, for any lazy state being held on the local CPU. */
+void sync_local_execstate(void);
+
 /*
  * Called by the scheduler to switch to another VCPU. This function must
  * call context_saved(@prev) when the local CPU is no longer running in
@@ -508,16 +563,13 @@ void startup_cpu_idle_loop(void);
  */
 unsigned long hypercall_create_continuation(
     unsigned int op, const char *format, ...);
+void hypercall_cancel_continuation(void);
 
 #define hypercall_preempt_check() (unlikely(    \
         softirq_pending(smp_processor_id()) |   \
         local_events_need_delivery()            \
     ))
 
-/* Protect updates/reads (resp.) of domain_list and domain_hash. */
-extern spinlock_t domlist_update_lock;
-extern rcu_read_lock_t domlist_read_lock;
-
 extern struct domain *domain_list;
 
 /* Caller must hold the domlist_read_lock or domlist_update_lock. */
@@ -527,7 +579,7 @@ extern struct domain *domain_list;
        (_d) = rcu_dereference((_d)->next_in_list )) \
 
 #define for_each_vcpu(_d,_v)                    \
- for ( (_v) = (_d)->vcpu[0];                    \
+ for ( (_v) = (_d)->vcpu ? (_d)->vcpu[0] : NULL; \
        (_v) != NULL;                            \
        (_v) = (_v)->next_in_list )
 
@@ -546,6 +598,9 @@ extern struct domain *domain_list;
  /* VCPU affinity has changed: migrating to a new CPU. */
 #define _VPF_migrating       3
 #define VPF_migrating        (1UL<<_VPF_migrating)
+ /* VCPU is blocked on memory-event ring. */
+#define _VPF_mem_event       4
+#define VPF_mem_event        (1UL<<_VPF_mem_event)
 
 static inline int vcpu_runnable(struct vcpu *v)
 {
@@ -564,18 +619,31 @@ void domain_pause_by_systemcontroller(struct domain *d);
 void domain_unpause_by_systemcontroller(struct domain *d);
 void cpu_init(void);
 
+struct scheduler;
+
+struct scheduler *scheduler_get_default(void);
+struct scheduler *scheduler_alloc(unsigned int sched_id, int *perr);
+void scheduler_free(struct scheduler *sched);
+int schedule_cpu_switch(unsigned int cpu, struct cpupool *c);
 void vcpu_force_reschedule(struct vcpu *v);
-void cpu_disable_scheduler(void);
+int cpu_disable_scheduler(unsigned int cpu);
 int vcpu_set_affinity(struct vcpu *v, cpumask_t *affinity);
-int vcpu_lock_affinity(struct vcpu *v, cpumask_t *affinity);
-int vcpu_locked_change_affinity(struct vcpu *v, cpumask_t *affinity);
-void vcpu_unlock_affinity(struct vcpu *v, cpumask_t *affinity);
 
 void vcpu_runstate_get(struct vcpu *v, struct vcpu_runstate_info *runstate);
 void domain_runstate_get(struct domain *d, domain_runstate_info_t *runstate);
 
 uint64_t get_cpu_idle_time(unsigned int cpu);
 
+/*
+ * Used by idle loop to decide whether there is work to do:
+ *  (1) Run softirqs; or (2) Play dead; or (3) Run tasklets.
+ */
+#define cpu_is_haltable(cpu)                    \
+    (!softirq_pending(cpu) &&                   \
+     cpu_online(cpu) &&                         \
+     !per_cpu(tasklet_work_to_do, cpu))
+
+void watchdog_domain_init(struct domain *d);
 void watchdog_domain_destroy(struct domain *d);
 
 #define IS_PRIV(_d) ((_d)->is_privileged)
@@ -585,17 +653,28 @@ void watchdog_domain_destroy(struct domain *d);
 
 #define is_hvm_domain(d) ((d)->is_hvm)
 #define is_hvm_vcpu(v)   (is_hvm_domain(v->domain))
-#define need_iommu(d)    ((d)->need_iommu && !(d)->is_hvm)
+#define is_pinned_vcpu(v) ((v)->domain->is_pinned || \
+                           cpus_weight((v)->cpu_affinity) == 1)
+#define need_iommu(d)    ((d)->need_iommu)
 
 void set_vcpu_migration_delay(unsigned int delay);
 unsigned int get_vcpu_migration_delay(void);
 
-extern int sched_smt_power_savings;
+extern bool_t sched_smt_power_savings;
 
 extern enum cpufreq_controller {
     FREQCTL_none, FREQCTL_dom0_kernel, FREQCTL_xen
 } cpufreq_controller;
 
+#define CPUPOOLID_NONE    -1
+
+struct cpupool *cpupool_get_by_id(int poolid);
+void cpupool_put(struct cpupool *pool);
+int cpupool_add_domain(struct domain *d, int poolid);
+void cpupool_rm_domain(struct domain *d);
+int cpupool_do_sysctl(struct xen_sysctl_cpupool_op *op);
+#define num_cpupool_cpus(c) (cpus_weight((c)->cpu_valid))
+
 #endif /* __SCHED_H__ */
 
 /*
